import os
import json
import vmprof
import pytest
from zipfile import ZipFile
from vmprof.reader import JittedCode, AssemblerCode
from vmprofconvert import convert
from vmprofconvert import convert_vmprof, convert_stats_with_pypylog
from vmprofconvert import convert_stats
from vmprofconvert import Converter
from vmprofconvert import Thread
from vmprofconvert import CATEGORY_PYTHON, CATEGORY_NATIVE, CATEGORY_JIT, CATEGORY_ASM, CATEGORY_JIT_INLINED, CATEGORY_MIXED, CATEGORY_INTERPRETER, CATEGORY_GC
from vmprofconvert.pypylog import parse_pypylog, cut_pypylog, rescale_pypylog, filter_top_level_logs
from vmprofconvert.__main__ import write_file_dict, save_zip, load_zip_dict, extract_files

class Dummystats():
    def __init__(self, profiles):
        self.profiles = profiles
        self.profile_lines = True
        self.profile_memory = False
        self.end_time = Dummytime(10)
        self.start_time = Dummytime(0)
        self.meta = {}
    
    def get_addr_info(self, addr):
        return ("py", self.addr_dict[addr], 0, "dummyfile.py")

class Dummytime():
    def __init__(self, time):
        self.time = time
    
    def timestamp(self):
        return self.time
    
def test_example():
    path = os.path.join(os.path.dirname(__file__), "profiles/example.prof")
    result = convert(path)
    
def test_stringarray():
    t = Thread()
    index = t.add_string("Hallo")
    assert index == 0
    index = t.add_string("Hallo")
    assert index == 0
    index = t.add_string("Huhu")
    assert index == 1
    assert t.stringarray == ["Hallo", "Huhu"]

def test_stacktable():
    t = Thread()
    assert t.add_stack([], 0) is None
    stackindex0 = t.add_stack([1,2,3], [CATEGORY_PYTHON, CATEGORY_PYTHON, CATEGORY_PYTHON])# Top of Stack is 3    [stack], [categories]
    stackindex1 = t.add_stack([1,2,3], [CATEGORY_PYTHON, CATEGORY_PYTHON, CATEGORY_PYTHON])
    assert stackindex0 == stackindex1 == 2
    assert t.stacktable == [[1,None, CATEGORY_PYTHON], [2,0, CATEGORY_PYTHON], [3,1, CATEGORY_PYTHON]]
    stackindex2 = t.add_stack([1,2,3,4], [CATEGORY_PYTHON, CATEGORY_PYTHON, CATEGORY_PYTHON, CATEGORY_NATIVE])
    assert stackindex2 == stackindex1 + 1

def test_frametable():
    t = Thread()
    frameindex0 = t.add_frame("duck", -1, "dummyfile.py", CATEGORY_PYTHON, -1, -1)# string, line, file, nativesymbol_index, addr
    frameindex1 = t.add_frame("duck", -1, "dummyfile.py", CATEGORY_PYTHON, -1, -1)
    assert frameindex0 == frameindex1 == 0
    frameindex2 = t.add_frame("goose", -1, "dummyfile.py", CATEGORY_PYTHON, -1, -1)
    assert frameindex2 == frameindex1 + 1
    assert t.frametable == [[0, -1, -1, 0],[1, -1, -1, 0]]
    assert t.stringarray == ["duck", "dummyfile.py" , "goose"]

def test_functable():
    t = Thread()
    funcindex0 = t.add_func("function_a", "dummyfile.py", 7, CATEGORY_PYTHON, -1)# func, file, line, category, resource, is_python
    funcindex1 = t.add_func("function_a", "dummyfile.py", 7, CATEGORY_PYTHON, -1)
    assert funcindex0 == funcindex1 == 0
    funcindex2 = t.add_func("function_b", "dummyfile.py", 17, CATEGORY_PYTHON, -1)
    assert funcindex2 == funcindex1 + 1
    assert t.functable == [[0, 1, 7, -1, True],[2, 1, 17, -1, True]]
    assert t.stringarray == ["function_a", "dummyfile.py" , "function_b"]

def test_sampleslist():
    t = Thread()
    t.add_sample(0, 7)# samples with same stack
    t.add_sample(0, 13)# samples with same stack
    t.add_sample(1, 17)
    assert t.samples == [[0, 7], [0, 13], [1, 17]]

def test_walksamples():
    c = Converter()
    vmprof_like_sample0 = (
        [0,# address
         -7,# line
         1,
         -17], # callstack with line numbers
         1, # samples count
         12345, #thread id
         0 # memory usage in kb
    )
    vmprof_like_sample1 = (
        [0,
         -7,
         2,
         -117], # callstack with line numbers
         1, # samples count
         12345, #thread id
         0 # memory usage in kb
    )
    stats = Dummystats([vmprof_like_sample0, vmprof_like_sample1])
    stats.addr_dict = {0: "function_a", 1: "function_b", 2: "function_c"}
    c.walk_samples(stats)
    t = c.threads[12345] # info now stored in thread inside Converter
    assert t.stringarray == ["function_a", "dummyfile.py", "function_b", "function_c"]
    assert t.frametable == [[0, 0, 7, 0], [1, 1, 17, 0], [2, 2, 117, 0]]# stringtableindex, nativesymbol_index, line, category
    assert t.stacktable == [[0, None, CATEGORY_PYTHON], [1, 0, CATEGORY_PYTHON], [2, 0, CATEGORY_PYTHON]]
    assert t.samples == [[1, 0.0], [2, 5000.0]]# stackindex time dummyeventdelay = 7

def test_walksample_vmprof():
    path = os.path.join(os.path.dirname(__file__), "profiles/example.prof")
    c = convert_vmprof(path)
    t = list(c.threads.values())[0]# get thread from Converter
    assert len(t.samples) == 2535# number of samples in example.prof

def test_dumps_simple_profile():
    c = Converter()
    vmprof_like_sample0 = (
        [0,# address
         -7,# line
         1,
         -17], # callstack with line numbers
         1, # samples count
         12345, #thread id
         0 # memory usage in kb
    )
    vmprof_like_sample1 = (
        [0,
         -7,
         1,
         -17,
         2,
         -27], # callstack with line numbers
         1, # samples count
         12345, #thread id
         0 # memory usage in kb
    )
    stats = Dummystats([vmprof_like_sample0, vmprof_like_sample1])
    stats.addr_dict = {0: "function_a", 1: "function_b", 2: "function_c"}
    c.walk_samples(stats)
    jsonstr = c.dumps_static()
    path = os.path.join(os.path.dirname(__file__), "profiles/example.json")
    with open(path, "w") as output_file:
        output_file.write(json.dumps(json.loads(jsonstr), indent=2))
    jsonobject = json.loads(jsonstr)
    dumped_funcreferences = jsonobject["threads"][0]["frameTable"]["func"]
    expected_dumped_funcreferences = [0,1,2]
    assert dumped_funcreferences == expected_dumped_funcreferences

def test_dumps_vmprof_no_lines():
    path = os.path.join(os.path.dirname(__file__), "profiles/example.prof")#profile_lines = False
    c = convert(path)
    jsonstr = c.dumps_static()
    path = os.path.join(os.path.dirname(__file__), "profiles/example.json")
    jsonobject = json.loads(jsonstr)
    dumped_frametable = jsonobject["threads"][0]["frameTable"]
    assert dumped_frametable["line"][0] == 172
    assert dumped_frametable["line"][1] == 64
    
def test_dumps_vmprof():
    path = os.path.join(os.path.dirname(__file__), "profiles/example.prof")
    c = convert(path)
    jsonstr = c.dumps_static()
    path = os.path.join(os.path.dirname(__file__), "profiles/example.json")
    with open(path, "w") as output_file:
        output_file.write(json.dumps(json.loads(jsonstr), indent=2))
    t = list(c.threads.values())[0]
    assert "function_a" in str(t.stringarray)


def test_dump_vmprof_meta():
    path = os.path.join(os.path.dirname(__file__), "profiles/example.prof")
    stats = vmprof.read_profile(path)
    c = Converter()
    meta = c.dump_vmprof_meta(stats)
    assert meta["interval"] == 0.000194
    assert meta["startTime"] == 1681890179831.0
    assert meta["shutdownTime"] == 1681890180325.0
    assert meta["platform"] == "Windows"
    assert meta["oscpu"] == "Windows 64bit"
    assert meta["abi"] == "cpython"# data from example.prof

def test_dumps_vmprof_with_meta():
    path = os.path.join(os.path.dirname(__file__), "profiles/vmprof_cpuburn.prof")
    jsonstr = convert_stats(path)
    path = os.path.join(os.path.dirname(__file__), "profiles/vmprof_cpuburn.json")
    with open(path, "w") as output_file:
        output_file.write(json.dumps(json.loads(jsonstr), indent=2))
    assert True

def test_profiles():
    #test multiple profiles in one run
    profiles = []
    profiles.append("profiles/example.prof")
    profiles.append("profiles/vmprof_cpuburn.prof")# cpuburn.py can be found in vmprof-python github repo
    profiles.append("profiles/multithread_example.prof")
    profiles.append("profiles/pypy-pystone.prof")
    profiles.append("profiles/example_with_lines.prof")
    expected_samples_count = {}
    expected_samples_count["profiles/example.prof"] = 2535
    expected_samples_count["profiles/vmprof_cpuburn.prof"] = 5551
    expected_samples_count["profiles/multithread_example.prof"] = 27688
    expected_samples_count["profiles/pypy-pystone.prof"] = 375
    expected_samples_count["profiles/example_with_lines.prof"] = 3431
    for profile in profiles:
        path = os.path.join(os.path.dirname(__file__), profile)
        c = convert_vmprof(path)
        threads = list(c.threads.values())
        samples_c = sum(len(thread.samples) for thread in threads)
        assert samples_c == expected_samples_count[profile]

def test_multiple_threads():
    c = Converter()
    vmprof_like_sample0 = (
        [0,# address
        -7,# line
        1,
        -17], # callstack with line numbers
        1, # samples count
        12345, #thread id
        0 # memory usage in kb
    )
    vmprof_like_sample1 = (
        [0,
        -7,
        1,
        -17,
        2,
        -27], # callstack with line numbers
        1, # samples count
        54321, #thread id
        0 # memory usage in kb
    )
    stats = Dummystats([vmprof_like_sample0, vmprof_like_sample1])
    stats.addr_dict = {0: "function_a", 1: "function_b", 2: "function_c"}
    c.walk_samples(stats)
    path = os.path.join(os.path.dirname(__file__), "profiles/dummy.json")
    with open(path, "w") as output_file:
        output_file.write(json.dumps(json.loads(c.dumps_static()), indent=2))
    t_12345 = c.threads[12345]
    t_54321 = c.threads[54321]
    threadids = list(c.threads.keys())
    expected_threadids = [12345, 54321]
    assert threadids == expected_threadids
    assert len(t_12345.stringarray) == 3
    assert len(t_54321.stringarray) == 4

def test_dumps_vmprof_memory():
    path = os.path.join(os.path.dirname(__file__), "profiles/vmprof_cpuburn.prof")
    jsonstr = convert_stats(path)
    profile = json.loads(jsonstr)
    memory_samples = profile["counters"][0]["sampleGroups"][0]["samples"]
    assert memory_samples["count"][0] == 11972000
    assert memory_samples["count"][1] == 11972000 # Firefox wont show memory unless two samples are not zero
    assert sum(memory_samples["count"][2:]) == 0
    assert len(memory_samples["count"]) == 5551

def test_dumps_filename_lines():
    path = os.path.join(os.path.dirname(__file__), "profiles/vmprof_cpuburn.prof")
    jsonstr = convert_stats(path)
    path = os.path.join(os.path.dirname(__file__), "profiles/vmprof_cpuburn.json")
    profile = json.loads(jsonstr)
    stringarray = profile["threads"][0]["stringArray"]
    functable = profile["threads"][0]["funcTable"] 
    native_file_index =  stringarray.index("-")
    assert native_file_index not in functable["fileName"]

def test_jit_asm_inline():
    c = Converter()
    vmprof_like_sample0 = (
        [0,
         0,# JIT/ASM/Native Frames have line == 0 in Dummystats
        JittedCode(0),# address
        7,# line
        JittedCode(1),# address
        17,# line
        AssemblerCode(2),
        -117], # callstack with line numbers
        1, # samples count
        12345, #thread id
        0 # memory usage in kb
    )
    stats = Dummystats([vmprof_like_sample0])
    stats.addr_dict = {0: "function_a", 1: "function_b", 2: "function_c"}
    c.walk_samples(stats)
    path = os.path.join(os.path.dirname(__file__), "profiles/dummy.json")
    with open(path, "w") as output_file:
        output_file.write(json.dumps(json.loads(c.dumps_static()), indent=2))
    thread = c.threads[12345]
    assert thread.stacktable == [[1, None, CATEGORY_MIXED], [2, 0, CATEGORY_JIT_INLINED]]

def test_pypy_pystone():
    path = os.path.join(os.path.dirname(__file__), "profiles/pypy-pystone.prof")
    jsonstr = convert_stats(path)
    path = os.path.join(os.path.dirname(__file__), "profiles/pypy-pystone.json")
    with open(path, "w") as output_file:
        output_file.write(json.dumps(json.loads(jsonstr), indent=2))
    profile = json.loads(jsonstr)
    stacktable = profile["threads"][0]["stackTable"]     
    assert CATEGORY_JIT not in stacktable["category"]
    assert CATEGORY_ASM not in stacktable["category"]

def test_check_asm_frame():
    categories = []
    c = Converter()
    thread = Thread()
    stack_info = "asm_function"
    c.check_asm_frame(categories, stack_info, thread, None)
    assert categories == []# asm frames currently disabled 
    categories.append(CATEGORY_JIT)
    thread.add_frame("jit_function", 7, "dummyfile.py", CATEGORY_JIT, 0, -1)
    c.check_asm_frame(categories, stack_info, thread, 0)
    assert categories == [CATEGORY_JIT_INLINED]# jit frame + asm frame => jit_inlined frame

def test_add_native_frame():
    c = Converter()
    thread = Thread()
    stack_info = "native_function"
    c.add_native_frame(thread, stack_info)
    assert thread.frametable == [[0, -1, -1, 2]]
    assert thread.functable == [[0, 1, -1, -1, False]]
    assert thread.stringarray == [stack_info, ""]

def test_add_jit_frame_to_mixed():
    c = Converter()
    thread = Thread()
    categories =  [CATEGORY_PYTHON]
    addr_info_jit = ("", "function_a", 7, "dummyfile.py")
    frame_index0 = thread.add_frame(addr_info_jit[1], 7, addr_info_jit[3], CATEGORY_PYTHON, -1, -1)
    frames = [frame_index0]
    frame_index1 = c.add_jit_frame(thread, categories, addr_info_jit, frames)
    frames.append(frame_index1)
    assert categories == [CATEGORY_MIXED]
    assert frames == [1]

def test_add_jit_frame_not_mixed():
    c = Converter()
    thread = Thread()
    categories =  []
    frames = []
    addr_info_jit0 = ("", "function_a", 7, "dummyfile.py")
    addr_info_jit1 = ("", "function_b", 17, "dummyfile.py")
    frame_index0 = c.add_jit_frame(thread, categories, addr_info_jit0, frames)
    frames.append(frame_index0)
    frame_index1 = c.add_jit_frame(thread, categories, addr_info_jit1, frames)
    frames.append(frame_index1)
    assert categories == [CATEGORY_JIT, CATEGORY_JIT]
    assert frames == [0,1]

def test_add_vmprof_frame():
    # def add_vmprof_frame(self, addr_info, thread, stack_info, lineprof, j):# native or python frame
    c = Converter()
    thread = Thread()
    addr_info_py = ("py", "function_a", 7, "dummyfile.py")
    addr_info_n = ("n", "function_b", 0, "dummyfile.py")
    stack_info = [0, -7, 1, 0]# addr, line, addr, line
    profile_lines = True
    c.add_vmprof_frame(addr_info_py, thread, stack_info, profile_lines, CATEGORY_PYTHON, 0)
    c.add_vmprof_frame(addr_info_n, thread, stack_info, profile_lines, CATEGORY_NATIVE, 2)
    assert thread.frametable == [[0, 0, 7, 0], [1, 1, 0, 2]]
    assert thread.functable == [[0, 1, 7, 0, True], [2, 1, 0, 1, False]]
    assert thread.stringarray == ["function_a", "dummyfile.py", "function_b"]

def test_add_lib():
    c = Converter()
    lib_index0 = c.add_lib("duck.py", "func_a")
    lib_index1 = c.add_lib("duck.py", "func_a")
    lib_index2 = c.add_lib("goose.py", "func_b")
    assert lib_index0 == lib_index1
    assert lib_index1 == lib_index2 - 1
    assert c.libs == [["duck.py","func_a"],["goose.py", "func_b"]]# Not in a thread

def test_add_native_symbol():
    t = Thread()
    nativesymbol_index0 = t.add_nativesymbol(0, "function_x", 7)
    nativesymbol_index1 = t.add_nativesymbol(1, "function_y", 17)
    assert nativesymbol_index0 == nativesymbol_index1 - 1
    assert t.nativesymbols == [[0, 0, 7],[1, 1, 17]]# libindex, stringindex, addr

def test_add_resource():
    t = Thread()
    lib_index0 = 0
    lib_index1 = 1
    string_index0 = t.add_string("function_x")
    string_index1 = t.add_string("function_y")
    resource_index0 = t.add_resource(lib_index0, string_index0)
    resource_index1 = t.add_resource(lib_index1, string_index1)
    assert resource_index0 == resource_index1 - 1
    assert t.resourcetable == [[0,0],[1,1]]# libindex, stringindex

def test_add_marker():
    t = Thread()
    t.add_marker(7, 17, 0) # starttime endtime stringtable_index
    t.add_marker(17, 117, 1)
    expected_marker0 = [7, 17, 0]
    expected_marker1 = [17, 117, 1]
    assert t.markers == [expected_marker0, expected_marker1]

def test_parse_pypylog():
    pypylog_path = os.path.join(os.path.dirname(__file__), "profiles/pystone.pypylog")
    pypylog = parse_pypylog(pypylog_path)
    assert pypylog[0] == [314906064138,"gc-set-nursery-size", True, 0]
    assert pypylog[-1] == [317567248367,"jit-summary", False, 0]
    assert len(pypylog) == 8248

def test_cut_pypylog():
    pypylog_path = os.path.join(os.path.dirname(__file__), "profiles/pystone.pypylog")
    initial_pypylog = parse_pypylog(pypylog_path)
    cutted_pypylog = cut_pypylog(initial_pypylog, 1000, 700)
    expected_length = int(8248 * 0.7)
    assert len(initial_pypylog) == 8248
    assert len(cutted_pypylog) == expected_length

def test_rescale_pypylog():
    pypylog_path = os.path.join(os.path.dirname(__file__), "profiles/pystone.pypylog")
    initial_pypylog = parse_pypylog(pypylog_path)
    rescaled_pypylog = rescale_pypylog(initial_pypylog[:1000], 10000000)
    assert rescaled_pypylog[7][0] == 70
    assert rescaled_pypylog[-1][0] == 9990

def test_filter_top_level_logs():
    pypylog_path = os.path.join(os.path.dirname(__file__), "profiles/pystone.pypylog")
    initial_pypylog = parse_pypylog(pypylog_path)
    filtered_pypylog = filter_top_level_logs(initial_pypylog[:25])
    assert initial_pypylog[6][1] != initial_pypylog[7][1]# nested action
    assert filtered_pypylog[6][1] == filtered_pypylog[7][1]# not nested action
    assert filtered_pypylog[6][2] != filtered_pypylog[7][2]# action start => action end

def test_create_pypylog_marker(): # create_pypylog_marker is no longer in use
    pypylog_path = os.path.join(os.path.dirname(__file__), "profiles/pystone.pypylog")
    t = Thread()
    initial_pypylog = parse_pypylog(pypylog_path)
    rescaled_pypylog = rescale_pypylog(initial_pypylog[:1000], 10000000)
    filtered_pypylog = filter_top_level_logs(rescaled_pypylog[:20])
    t.create_pypylog_marker(filtered_pypylog)# marker do not use categories, but strings 
    assert t.markers[0] == [0, 10, 1]#gc or jit
    assert t.markers[1] == [11, 19, 0]#interp
    assert t.markers[2] == [20, 30, 2]#gc or jit
    assert t.markers[3] == [31, 39, 0]#interp

def test_walk_pypylog_marker(): # create_pypylog_marker is no longer in use
    pypylog_path = os.path.join(os.path.dirname(__file__), "profiles/pystone.pypylog")
    converter = Converter()
    initial_pypylog = parse_pypylog(pypylog_path)
    rescaled_pypylog = rescale_pypylog(initial_pypylog[:100], 1000000)
    converter.walk_pypylog(rescaled_pypylog)
    thread = converter.threads[7]
    
    gc_set_nursery_size_id = thread.add_string("gc-set-nursery-size")# get string ids, since name is string not category
    gc_hardware_id = thread.add_string("gc-hardware")
    interpreter_id = thread.add_string("interpreter")

    expected_interval = 1000000/100  * 0.001

    assert thread.markers[0] == [0, expected_interval, gc_set_nursery_size_id] # first log
    assert thread.markers[1] == [expected_interval + 1,2 * expected_interval -1, interpreter_id]# interpreter
    assert thread.markers[2] == [2 * expected_interval, 3 * expected_interval, gc_hardware_id]# second log
    assert thread.markers[3] == [3 * expected_interval + 1,4 * expected_interval -1, interpreter_id]# interpreter
    assert thread.markers[4] == [4 * expected_interval, 5 * expected_interval, gc_hardware_id]# third log

def test_get_unused_tid():
    c = Converter()
    c.threads[2] = Thread()
    c.threads[7] = Thread()
    c.threads[17] = Thread()
    unused_tid = c.get_unused_tid()
    assert unused_tid == 1

def test_add_pypylog_sample():
    c = Converter()
    thread = Thread()
    logname = "gc_dummy"
    logtime_start = 7
    logtime_end = 17
    c.add_pypylog_sample(thread, logname, logtime_start, logtime_end)
    samples = thread.samples
    assert samples[0] == [0, 7]
    assert samples[1] == [0, 17]

def test_add_pypylog_interp_sample():
    c = Converter()
    thread = Thread()
    logtime_end = 17
    next_logtime_start = 117
    c.add_pypylog_interp_sample(thread, logtime_end, next_logtime_start)
    samples = thread.samples
    stacktable = thread.stacktable
    assert samples[0] == [0, 17]
    assert samples[1] == [0, 117]
    assert stacktable[0] == [0, None, CATEGORY_INTERPRETER]

def test_walk_pypylog():
    c = Converter()
    test_pypylog = [
        [7, "gc_example_action_a", True, 0],
        [17, "gc_example_action_a", False, 0]
    ]
    c.walk_pypylog(test_pypylog)
    thread = c.threads[7]
    stringarray = thread.stringarray
    assert stringarray[0] == "gc_example_action_a"


def test_walk_pypylog_interpreter_samples():
    c = Converter()
    test_pypylog = [# this just represents the order in time time. Order in samples list differs: gc gc interp gc gc interp
        [0, "gc_example_action_a", True, 0],
        [5, "gc_example_action_a", False, 0],
        # interpreter sample here
        [10, "gc_example_action_a", True, 0],
        [15, "gc_example_action_a", False, 0],
        # no interpreter sample here ### maybe we dont need this case
        [16, "gc_example_action_a", True, 0],
        [17, "gc_example_action_a", False, 0],
        # interpreter sample here
        [25, "gc_example_action_a", True, 0],
        [30, "gc_example_action_a", False, 0]
    ]
    c.walk_pypylog(test_pypylog)
    thread = c.threads[7]

    samples = thread.samples
    stacktable = thread.stacktable

    sample_0 = samples[0]# should be gc_example_action_a
    sample_2 = samples[2]# should be gc_example_action_a
    sample_4 = samples[4]# should be interpreter
    sample_6 = samples[6]# should be gc_example_action_a
    sample_8 = samples[8]# should be gc_example_action_a
    sample_10 = samples[10]# should be interpreter

    stack_0 = stacktable[sample_0[0]]
    stack_2 = stacktable[sample_2[0]]
    stack_4 = stacktable[sample_4[0]]
    stack_6 = stacktable[sample_6[0]]
    stack_8 = stacktable[sample_8[0]]
    stack_10 = stacktable[sample_10[0]]

    assert stack_0[2] == CATEGORY_GC
    assert stack_2[2] == CATEGORY_INTERPRETER
    assert stack_4[2] == CATEGORY_GC
    assert stack_6[2] == CATEGORY_GC
    assert stack_8[2] == CATEGORY_INTERPRETER
    assert stack_10[2] == CATEGORY_GC
    assert len(samples) == 12

def test_walk_full_pypylog():
    c = Converter()
    test_pypylog = [
        [0, "gc_example_action_a", True, 0],
        [1, "gc_example_action_b", True, 1],
        [2, "gc_example_action_c", True, 2],
        [3, "gc_example_action_c", False, 2], # top => two samples created
        [4, "gc_example_action_b", False, 1], # not top
        [5, "gc_example_action_a", False, 0], # not top
        [6, "gc_example_action_a", True, 0],
        [7, "gc_example_action_a", False, 0] # top => two samples created
    ]
    c.walk_pypylog(test_pypylog)
    thread = c.threads[7]
    samples = thread.samples
    stacktable = thread.stacktable
    frametable = thread.frametable
    functable = thread.functable
    stringarray = thread.stringarray

    sample_0 = samples[0]
    sample_2 = samples[2]
    sample_4 = samples[4]

    stack_0 = stacktable[sample_0[0]]
    stack_2 = stacktable[sample_2[0]]
    stack_4 = stacktable[sample_4[0]]

    frame_0 = frametable[stack_0[0]]
    frame_2 = frametable[stack_2[0]]
    frame_4 = frametable[stack_4[0]]

    func_0 = functable[frame_0[0]]
    func_2 = functable[frame_2[0]]
    func_4 = functable[frame_4[0]]

    funcname_0 = stringarray[func_0[0]]# gc_example_action_c
    funcname_2 = stringarray[func_2[0]]# interpreter
    funcname_4 = stringarray[func_4[0]]# gc_example_action_a

    assert funcname_0 == "gc_example_action_c" # action c is a top level action
    assert funcname_2 == "interp" # interpreter (why did i call this 'interp'?)
    assert funcname_4 == "gc_example_action_a" # action a is a top level action, but c should be at index 0

    assert len(samples) == 6# there should be only four samples 2x action c, 2x action a and now 2x interpreter
    
def test_dumps_vmprof_without_pypylog():
    vmprof_path = os.path.join(os.path.dirname(__file__), "profiles/vmprof_cpuburn.prof")
    pypylog_path = None
    times = None
    jsonstr, _ = convert_stats_with_pypylog(vmprof_path, pypylog_path, times)
    profile = json.loads(jsonstr)
    samples = profile["threads"][0]["samples"] 
    markers = profile["threads"][0]["markers"]
    assert len(samples["stack"]) == 5551
    assert markers["data"] == []

def test_dumps_vmprof_with_pypylog(): ### TODO: re-enable assert as soon as interpreter frames can be created again
    vmprof_path = os.path.join(os.path.dirname(__file__), "profiles/vmprof_cpuburn.prof")
    pypylog_path = os.path.join(os.path.dirname(__file__), "profiles/pystone.pypylog")
    times = (0, 42.368387)
    jsonstr, _ = convert_stats_with_pypylog(vmprof_path, pypylog_path, times)
    profile = json.loads(jsonstr)
    samples = profile["threads"][0]["samples"] 
    stringarray = profile["threads"][1]["stringArray"]
    assert len(samples["stack"]) == 5551
    #assert stringarray[0] == "interpreter"
    assert stringarray[0] == "gc-set-nursery-size"

def test_write_file_dict():
    file_dict = {
        "file_a": "duck.py",
        "file_b": "goose.prof"
    }
    zip_path = "tmpzip/examplezip.zip"
    dict_path = None
    json_dict = None
    
    os.mkdir("tmpzip")

    with ZipFile(zip_path, "w") as examplezip:# create zip file and write file_dict
        write_file_dict(file_dict, examplezip)

    with ZipFile(zip_path, "r") as examplezip:
        dict_path = examplezip.extract("dict.json")# extract file_dict

    with open(dict_path, "r") as exampledict:# load file_dict
        json_dict = json.loads(exampledict.read())

    os.remove(zip_path)# cleanup
    os.remove(dict_path)
    os.rmdir("tmpzip")

    assert not os.path.exists("tmp")
    assert json_dict["file_a"] == file_dict["file_a"]
    assert json_dict["file_b"] == file_dict["file_b"]

def test_save_zip():
    file_dict = {
        "duck.py": "duck.py",
        "/home/users/me/goose.prof": "goose.prof",
        "C:\\users\\myself\\frog.jitlog": "frog.jitlog"
    }

    with open(file_dict["duck.py"], "w") as file_a:# create dummy files
        file_a.write("print(\"quack\")")

    with open(file_dict["/home/users/me/goose.prof"], "w") as file_b:# create dummy files
        file_b.write("0x7")

    with open(file_dict["C:\\users\\myself\\frog.jitlog"], "w") as file_c:# create dummy files
        file_c.write("0x17")

    extract_folder = "extracted"
    zip_folder = "tmpzip"
    zip_path = zip_folder + "/examplezip.zip"
   
    os.mkdir(zip_folder)

    save_zip(zip_path, file_dict)# save dummy files in zip with dict

    os.remove(file_dict["duck.py"])# remove local dummy files
    os.remove(file_dict["/home/users/me/goose.prof"])
    os.remove(file_dict["C:\\users\\myself\\frog.jitlog"])

    os.mkdir(extract_folder)

    with ZipFile(zip_path, "r") as examplezip:# open zip extract dict
        dict_path = examplezip.extract("dict.json", extract_folder)

    with open(dict_path, "r") as exampledict:# read file_dict
        json_dict = json.loads(exampledict.read())
    
    new_file_paths = {}
    with ZipFile(zip_path, "r") as examplezip:# open zip extract files listed in file_dict
        for path in list(json_dict.keys()):
            filename = json_dict[path]
            new_file_paths[path] = examplezip.extract(filename, extract_folder)
 
    filecontent = {}
    for path in list(new_file_paths.keys()):# load content of extracted files 
        with open(new_file_paths[path], "r") as file:
            filecontent[path] = file.read()

    os.remove(zip_path)# cleanup
    os.rmdir(zip_folder)
    os.remove(dict_path)
    for file in new_file_paths:
        os.remove(new_file_paths[file])
    os.rmdir(extract_folder)

    assert filecontent["duck.py"] == "print(\"quack\")"
    assert filecontent["/home/users/me/goose.prof"] == "0x7"
    assert filecontent["C:\\users\\myself\\frog.jitlog"] == "0x17"


def test_load_zip_dict():
    file_dict = {
        "duck.py": "duck.py",
        "/home/users/me/goose.prof": "goose.prof",
        "C:\\users\\myself\\frog.jitlog": "frog.jitlog"
    }

    zip_folder = "tmpzip"
    zip_path = os.path.join(zip_folder, "examplezip.zip")
   
    os.mkdir(zip_folder)

    with ZipFile(zip_path, "w") as examplezip:# create zip file and write file_dict
        write_file_dict(file_dict, examplezip)
    
    zip_dict = load_zip_dict(zip_path, zip_folder)

    os.remove(zip_path)# cleanup
    os.rmdir(zip_folder)

    assert zip_dict == file_dict
    
def test_extract_files():
    file_dict = {
        "duck.py": "duck.py",
        "/home/users/me/goose.prof": "goose.prof",
        "C:\\users\\myself\\frog.jitlog": "frog.jitlog"
    }

    with open(file_dict["duck.py"], "w") as file_a:# create dummy files
        file_a.write("print(\"quack\")")

    with open(file_dict["/home/users/me/goose.prof"], "w") as file_b:# create dummy files
        file_b.write("0x7")

    with open(file_dict["C:\\users\\myself\\frog.jitlog"], "w") as file_c:# create dummy files
        file_c.write("0x17")

    extract_folder = "extracted"
    zip_folder = "tmpzip"
    zip_path = zip_folder + "/examplezip.zip"
   
    os.mkdir(zip_folder)

    save_zip(zip_path, file_dict)# save dummy files in zip with dict

    os.remove(file_dict["duck.py"])# remove local dummy files
    os.remove(file_dict["/home/users/me/goose.prof"])
    os.remove(file_dict["C:\\users\\myself\\frog.jitlog"])

    zip_dict = load_zip_dict(zip_path, zip_folder)

    new_file_paths = extract_files(zip_dict, zip_path, extract_folder)

    filecontent = {}
    for path in list(new_file_paths.keys()):# load content of extracted files 
        with open(new_file_paths[path], "r") as file:
            filecontent[path] = file.read()

    os.remove(zip_path)# cleanup
    os.rmdir(zip_folder)
    for file in new_file_paths:
        os.remove(new_file_paths[file])
    os.rmdir(extract_folder)

    assert filecontent["duck.py"] == "print(\"quack\")"
    assert filecontent["/home/users/me/goose.prof"] == "0x7"
    assert filecontent["C:\\users\\myself\\frog.jitlog"] == "0x17"


def test_categories_in_frametable():
    ### Test if categories are stored in frameTable
    ### Firefox Profiler doesnt use stackTable categories anymore
    ### But im unsure if thats intended so we have redundant categories for now
    c = Converter()
    thread = Thread()
    addr_info_py = ("py", "function_a", 7, "dummyfile.py")
    addr_info_n = ("n", "function_b", 0, "dummyfile.py")
    addr_info_jit = ("jit", "function_c", 0, "dummyfile.py")
    stack_info = [0, -7, 1, 0]# addr, line, addr, line
    profile_lines = True

    frame_py = c.add_vmprof_frame(addr_info_py, thread, stack_info, profile_lines, CATEGORY_PYTHON, 0)
    frame_n  = c.add_vmprof_frame(addr_info_n, thread, stack_info, profile_lines, CATEGORY_NATIVE, 2)

    # categories & frames inside 'walk_samples' should look like this
    categories = [CATEGORY_PYTHON, CATEGORY_NATIVE]
    frames = [frame_py, frame_n]

    c.add_jit_frame(thread, categories, addr_info_jit, frames)

    assert thread.frametable == [[0, 0, 7, CATEGORY_PYTHON], [1, 1, 0, CATEGORY_NATIVE], [2, 2, 0, CATEGORY_JIT]] # categories now also in frametable
    assert thread.functable == [[0, 1, 7, 0, True], [2, 1, 0, 1, False], [3, 1, 0, 2, True]]
    assert thread.stringarray == ["function_a", "dummyfile.py", "function_b", "function_c"]

def test_dumps_vmprof_categories_inf_frametable():
    ### Test if frameTable categories are correctly dumped into json
    vmprof_path = os.path.join(os.path.dirname(__file__), "profiles/vmprof_cpuburn.prof")

    jsonstr, _ = convert_stats_with_pypylog(vmprof_path=vmprof_path, pypylog_path=None, times=None)
    profile = json.loads(jsonstr)
    frames = profile["threads"][0]["frameTable"] 

    # this profile contains python and native frames
    assert CATEGORY_PYTHON in frames["category"]
    assert CATEGORY_NATIVE in frames["category"]

# This test does only make sense when running vmprof with sample-timestamp support like 
# https://github.com/Cskorpion/vmprof-python/tree/sample_timestamps
# Test disabled until the vmprof main branch supports sample-timestamps
@pytest.mark.skip()
def test_timestamps():
    path = os.path.join(os.path.dirname(__file__), "profiles/cpuburn.cpython.tsprof")
    converter = convert_vmprof(path)

    thread = list(converter.threads.values())[0]

    samples = thread.samples

    assert 10.832965 - 0.00001 < samples[0][1] < 10.832965 + 0.00001 # different rounding
    # 451.370060965 - 451.359228 # sample timestamp - start timestamp
    assert 30430.785875 - 0.00001 < samples[-1][1] < 30430.785875 + 0.00001
    # 481.790013875 - 451.359228 



class InvalidFormatException(Exception):
    pass

def check_processed_profile(prof):
    if "meta" not in prof:
        raise InvalidFormatException("meta key missing")
    else:
        check_meta(prof["meta"])
    if "libs" not in prof:
        raise InvalidFormatException("libs key missing")
    if "pages" not in prof:
        raise InvalidFormatException("pages key missing")
    if "counters" in prof and len(prof["counters"]) != 0:
        check_counters(prof["counters"])
    if "threads" not in prof:
        raise InvalidFormatException("threads key missing")
    else:
        check_threads(prof["threads"])

def check_meta(meta):
    if "interval" not in meta:
        raise InvalidFormatException("interval key missing")
    if "startTime" not in meta:
        raise InvalidFormatException("startTime key missing")
    if "abi" not in meta:
        raise InvalidFormatException("abi key missing")
    if "oscpu" not in meta:
        raise InvalidFormatException("oscpu key missing")
    if "platform" not in meta:
        raise InvalidFormatException("platform key missing")
    if "processType" not in meta:
        raise InvalidFormatException("processType key missing")
    if "stackwalk" not in meta:
        raise InvalidFormatException("stackwalk key missing")
    if "debug" not in meta:
        raise InvalidFormatException("debug key missing")
    if "version" not in meta:
        raise InvalidFormatException("version key missing")
    if "categories" not in meta:
        raise InvalidFormatException("categories key missing")
    if "preprocessedProfileVersion" not in meta:
        raise InvalidFormatException("preprocessedProfileVersion key missing")
    if "symbolicated" not in meta:
        raise InvalidFormatException("symbolicated key missing")
    if "markerSchema" not in meta:
        raise InvalidFormatException("markerSchema key missing")
   

def check_counters(counters):
    for counter in counters:
        if "name" not in counter:
            raise InvalidFormatException("name key missing")
        if "category" not in counter:
            raise InvalidFormatException("category key missing")
        if "description" not in counter:
            raise InvalidFormatException("description key missing")
        if "pid" not in counter:
            raise InvalidFormatException("pid key missing")
        if "mainThreadIndex" not in counter:
            raise InvalidFormatException("mainThreadIndex key missing")
        if "sampleGroups" not in counter:
            raise InvalidFormatException("sampleGroups key missing")
        else:
            check_samplegroups(counter["sampleGroups"])

def check_samplegroups(samplegroups):
    for samplegroup in samplegroups:
        if "id" not in samplegroup:
            raise InvalidFormatException("sampleGroups id key missing")
        if "samples" not in samplegroup:
            raise InvalidFormatException("sampleGroups samples key missing")
        else:
            if "length" not in samplegroup["samples"]:
                raise InvalidFormatException("sampleGroups length key missing")
            if "time" not in samplegroup["samples"]:
                raise InvalidFormatException("sampleGroups time key missing")
            if "count" not in samplegroup["samples"]:
                raise InvalidFormatException("sampleGroups count key missing")
        
def check_threads(threads):
    for thread in threads:
        if "name" not in thread:
            raise InvalidFormatException("name key missing")
        if "isMainThread" not in thread:
            raise InvalidFormatException("isMainThread key missing")
        if "processType" not in thread:
            raise InvalidFormatException("processType key missing")
        if "processName" not in thread:
            raise InvalidFormatException("processName key missing")
        if "processStartupTime" not in thread:
            raise InvalidFormatException("processStartupTime key missing")
        if "processShutdownTime" not in thread:
            raise InvalidFormatException("processShutdownTime key missing")
        if "registerTime" not in thread:
            raise InvalidFormatException("registerTime key missing")
        if "unregisterTime" not in thread:
            raise InvalidFormatException("unregisterTime key missing")
        if "tid" not in thread:
            raise InvalidFormatException("tid key missing")
        if "pid" not in thread:
            raise InvalidFormatException("pid key missing")
        if "markers" not in thread:
            raise InvalidFormatException("markers key missing")
        else:
            check_markers(thread["markers"])
        if "frameTable" not in thread:
            raise InvalidFormatException("frameTable key missing")
        else:
            check_frametable(thread["frameTable"])
        if "funcTable" not in thread:
            raise InvalidFormatException("funcTable key missing")
        else:
            check_functable(thread["funcTable"])
        if "resourceTable" not in thread:
            raise InvalidFormatException("resourceTable key missing")
        else:
            check_resourcetable(thread["resourceTable"])
        if "stackTable" not in thread:
            raise InvalidFormatException("stackTable key missing")
        else:
            check_stacktable(thread["stackTable"])
        if "samples" not in thread:
            raise InvalidFormatException("samples key missing")
        else:
            check_samples(thread["samples"])
        if "stringArray" not in thread:
            raise InvalidFormatException("stringArray key missing")
        
def check_markers(markers):
    if "data" not in markers:
        raise InvalidFormatException("markers type key missing")
    if "length" not in markers:
        raise InvalidFormatException("markers length key missing")
        
def check_frametable(frametable):
    if "address" not in frametable:
        raise InvalidFormatException("frametable address key missing")
    if "inlineDepth" not in frametable:
        raise InvalidFormatException("frametable inlineDepth key missing")
    if "category" not in frametable:
        raise InvalidFormatException("frametable category key missing")
    if "subcategory" not in frametable:
        raise InvalidFormatException("frametable subcategory key missing")
    if "func" not in frametable:
        raise InvalidFormatException("frametable func key missing")
    if "innerWindowID" not in frametable:
        raise InvalidFormatException("frametable innerWindowID key missing")
    if "length" not in frametable:
        raise InvalidFormatException("frametable length key missing")
        
def check_functable(functable):
    if "isJS" not in functable:
        raise InvalidFormatException("functable isJS key missing")
    if "relevantForJS" not in functable:
        raise InvalidFormatException("functable relevantForJS key missing")
    if "name" not in functable:
        raise InvalidFormatException("functable name key missing")
    if "resource" not in functable:
        raise InvalidFormatException("functable resource key missing")
    if "fileName" not in functable:
        raise InvalidFormatException("functable fileName key missing")
    if "lineNumber" not in functable:
        raise InvalidFormatException("functable lineNumber key missing")
    if "columnNumber" not in functable:
        raise InvalidFormatException("functable columnNumber key missing")
    if "length" not in functable:
        raise InvalidFormatException("functable length key missing")
    
def check_resourcetable(resourcetable):
    if "type" not in resourcetable:
        raise InvalidFormatException("resourcetable type key missing")
    if "length" not in resourcetable:
        raise InvalidFormatException("resourcetable length key missing")

def check_stacktable(stacktable):
    if "frame" not in stacktable:
        raise InvalidFormatException("stacktable frame key missing")
    if "category" not in stacktable:
        raise InvalidFormatException("stacktable category key missing")
    if "subcategory" not in stacktable:
        raise InvalidFormatException("stacktable subcategory key missing")
    if "prefix" not in stacktable:
        raise InvalidFormatException("stacktable prefix key missing")
    if "length" not in stacktable:
        raise InvalidFormatException("stacktable length key missing")
    
def check_samples(samples):
    if "stack" not in samples:
        raise InvalidFormatException("samples stack key missing")
    if "time" not in samples:
        raise InvalidFormatException("samples time key missing")
    if "eventDelay" not in samples:
        raise InvalidFormatException("samples eventDelay key missing")
    if "weightType" not in samples:
        raise InvalidFormatException("samples weightType key missing")
    if "weight" not in samples:
        raise InvalidFormatException("samples weight key missing")
    if "length" not in samples:
        raise InvalidFormatException("samples length key missing")


import re

def parse_pypylog(path):
    raw_log = None
    if path:
        with open(path, "r") as ppl:
            raw_log = ppl.readlines()
    log = []
    if raw_log:
        # From gccauses.py
        depth = 0
        start = re.compile(r"\[([0-9a-fA-F]+)\] \{([\w-]+)")
        stop  = re.compile(r"\[([0-9a-fA-F]+)\] ([\w-]+)\}")
        for line in raw_log:
            match = start.match(line)
            starting = True
            if not match:
                match = stop.match(line)
                starting = False
            if match:
                timestamp = int(match.group(1), base=16)
                action = match.group(2)
                if starting:# not nice
                    log.append([timestamp, action, starting, depth])
                    depth += 1
                else:
                    depth -= 1
                    log.append([timestamp, action, starting, depth])
    return log

def cut_pypylog(pypylog, total_runtime_micros, vmprof_runtime_micros): 
    factor = vmprof_runtime_micros/total_runtime_micros
    limit = int(factor * len(pypylog))
    return pypylog[:limit]

def rescale_pypylog(pypylog, vmprof_runtime_micros):
    scaled_pypylog = []
    time = vmprof_runtime_micros / len(pypylog) * 0.001
    for i in range(len(pypylog)):
        line = pypylog[i]
        log_time = int(time * i)# len(pypylog) > micros
        scaled_pypylog.append([log_time, line[1], line[2], line[3]])
    return scaled_pypylog

def filter_top_level_logs(pypylog):
    filtered_pypylog = list(filter(lambda e: e[3] == 0, pypylog))
    return filtered_pypylog


class InvalidFormatException(Exception):
    pass

def check_gecko_profile(prof):
    if "meta" not in prof:
        raise InvalidFormatException("meta key missing")
    else:
        check_meta(prof["meta"])
    if "pages" not in prof:
        raise InvalidFormatException("pages key missing")
    if "libs" not in prof:
        raise InvalidFormatException("libs key missing")
    if "pausedRanges" not in prof:
        raise InvalidFormatException("pausedRanges key missing")
    if "threads" not in prof:
        raise InvalidFormatException("threads key missing")
    else:
        check_threads(prof["threads"])
    if "processes" not in prof:
        raise InvalidFormatException("processes key missing")

def check_meta(meta):
    if "version" not in meta:
        raise InvalidFormatException("version key missing")
    if "interval" not in meta:
        raise InvalidFormatException("interval key missing")
    if "stackwalk" not in meta:
        raise InvalidFormatException("stackwalk key missing")
    if "debug" not in meta:
        raise InvalidFormatException("debug key missing")
    if "startTime" not in meta:
        raise InvalidFormatException("startTime key missing")
    if "shutdownTime" not in meta:
        raise InvalidFormatException("stackwalk key missing")
    if "processType" not in meta:
        raise InvalidFormatException("processType key missing")
    if "platform" not in meta:
        raise InvalidFormatException("platform key missing")
    if "oscpu" not in meta:
        raise InvalidFormatException("oscpu key missing")
    if "abi" not in meta:
        raise InvalidFormatException("abi key missing")
    
def check_threads(threads):
    for thread in threads:
        if "name" not in thread:
            raise InvalidFormatException("name key missing")
        if "processType" not in thread:
            raise InvalidFormatException("processType key missing")
        if "processName" not in thread:
            raise InvalidFormatException("processName key missing")
        if "tid" not in thread:
            raise InvalidFormatException("tid key missing")
        if "pid" not in thread:
            raise InvalidFormatException("pid key missing")
        if "registerTime" not in thread:
            raise InvalidFormatException("registerTime key missing")
        if "unregisterTime" not in thread:
            raise InvalidFormatException("unregisterTime key missing")
        if "markers" not in thread:
            raise InvalidFormatException("markers key missing")
        else:
            check_markers(thread["markers"])
        if "samples" not in thread:
            raise InvalidFormatException("samples key missing")
        else:
            check_samples(thread["samples"])
        if "frameTable" not in thread:
            raise InvalidFormatException("frameTable key missing")
        else:
            check_frametable(thread["frameTable"])
        if "stackTable" not in thread:
            raise InvalidFormatException("stackTable key missing")
        else:
            check_stacktable(thread["stackTable"])
        if "stringTable" not in thread:
            raise InvalidFormatException("stringTable key missing")

def check_markers(markers):
    if "schema" not in markers:
        raise InvalidFormatException("schema key missing")
    else:
        check_marker_schema(markers["schema"])
    if "data" not in markers:
        raise InvalidFormatException("data key missing")
        
def check_marker_schema(schema):
    if "name" not in schema:
        raise InvalidFormatException("name key missing")
    if "time" not in schema:
        raise InvalidFormatException("time key missing")
    if "data" not in schema:
        raise InvalidFormatException("data key missing")
    
def check_samples(samples):
    if "schema" not in samples:
        raise InvalidFormatException("samples schema key missing")
    else:
        check_samples_schema(samples["schema"])
    if "data" not in samples:
        raise InvalidFormatException("samples data key missing")

def check_frametable(frametable):
    if "schema" not in frametable:
        raise InvalidFormatException("frameTable schema key missing")
    else:
        check_frametable_schema(frametable["schema"])
    if "data" not in frametable:
        raise InvalidFormatException("frameTable data key missing")

def check_stacktable(stacktable):
    if "schema" not in stacktable:
        raise InvalidFormatException("stackTable schema key missing")
    else:
        check_stacktable_schema(stacktable["schema"])
    if "data" not in stacktable:
        raise InvalidFormatException("stackTable data key missing")
    
def check_samples_schema(schema):
    if "stack" not in schema:
        raise InvalidFormatException("samples stack key missing")
    if "time" not in schema:
        raise InvalidFormatException("samples time key missing")
    if "eventDelay" not in schema:
        raise InvalidFormatException("samples eventDelay key missing")
    
def check_frametable_schema(schema):
    if "location" not in schema:
        raise InvalidFormatException("frameTable location key missing")
    if "relevantForJS" not in schema:
        raise InvalidFormatException("frameTable relevantForJS key missing")
    if "innerWindowID" not in schema:
        raise InvalidFormatException("frameTable innerWindowID key missing")
    if "implementation" not in schema:
        raise InvalidFormatException("frameTable implementation key missing")
    
def check_stacktable_schema(schema):
    if "frame" not in schema:
        raise InvalidFormatException("stackTable frame key missing")
    if "prefix" not in schema:
        raise InvalidFormatException("stackTable prefix key missing")


import sys
import json
import os
import webbrowser
import subprocess
import platform
import argparse
import time
import atexit
from zipfile import ZipFile
from vmprofconvert import convert_stats_with_pypylog
from symbolserver import start_server

def run_vmprof(path, argv, native, lines):
    profpath = path.replace(".py", ".prof")
    impl = platform.python_implementation()
    system = platform.system()
    args = " -o " + profpath

    env = os.environ.copy()

    jitlogpath = None
    pypylogpath = None
    if impl == "PyPy":
        args += " --jitlog"
        jitlogpath = profpath + ".jit"
        pypylogpath = path.replace(".py", ".pypylog")
        env["PYPYLOG"] = pypylogpath
    if impl == "CPython":
        if lines:
            args += " --lines"
        if system == "Linux" or system == "Darwin":
            args += " --mem"
    if not native:
        args += " --no-native"

    cmd_args = ""
    if len(argv) > 0:
        for arg in argv:
            cmd_args += " " + arg

    start_time = time.time()

    command = sys.executable + " -m vmprof" + args + " " + path + cmd_args
    subprocess.run(command, shell=True, env=env)

    end_time = time.time()

    return (profpath, jitlogpath, pypylogpath, (start_time, end_time))

def write_file_dict(file_dict, sharezip):
    #dump dict with real paths to new filenames mapping
    os.mkdir("tmp")
    with open("tmp/dict.json", "w") as dict_file:
        dict_file.write(json.dumps(file_dict, indent = 2))
    sharezip.write("tmp/dict.json", "dict.json")
    os.remove("tmp/dict.json")
    os.rmdir("tmp")
    pass

def save_zip(zip_path, path_dict):
    file_dict = {}
    if "times" in path_dict:
        file_dict["times"] = path_dict["times"]
        path_dict.pop("times")
    with ZipFile(zip_path, "w") as sharezip:
        ctr = 0
        for path in list(path_dict.keys()):
            new_filename = str(ctr) + "_" + os.path.basename(path_dict[path])
            sharezip.write(path_dict[path], new_filename)# new filename to prevent duplicate names
            file_dict[path] = new_filename
            ctr += 1
        write_file_dict(file_dict, sharezip)

def load_zip_dict(zip_path, folder):
    zip_dict = None
    dict_path = None
    with ZipFile(zip_path, "r") as inputzip:# open zip extract dict
        dict_path = inputzip.extract("dict.json", folder)
    if dict_path:
        with open(dict_path, "r") as file_dict:# read zip_dict
            zip_dict = json.loads(file_dict.read())
    os.remove(dict_path)
    return zip_dict

def extract_files(zip_dict, zip_path, folder):
    new_file_paths = {}
    with ZipFile(zip_path, "r") as inputzip:# open zip, extract files listed in zip_dict
        for path in list(zip_dict.keys()):
            filename = zip_dict[path]
            new_file_paths[path] = inputzip.extract(filename, folder)
    return new_file_paths

def cleanup(folder, path_dict):
    path_dict["json"] = path_dict["prof"] + ".json"# json is created, not loaded
    for path in path_dict:
        os.remove(path_dict[path])
    os.rmdir(folder)

def get_paths(path_dict):
    path = None
    jitlogpath = None
    pypylogpath = None
    if "prof" in path_dict:
        path = path_dict["prof"]
    if "jitlog" in path_dict:
        jitlogpath = path_dict["jitlog"]
    if "pypylog" in path_dict:
        pypylogpath = path_dict["pypylog"]
    return path, jitlogpath, pypylogpath


def main():
    parser = argparse.ArgumentParser(prog="vmprof-firefox-converter", description="convert vmprof profiles or run vmprof directly")
    parser.add_argument("-convert", metavar = "convert_file", dest = "convert_file", nargs = 1, help = "convert vmprof profile or zip")
    parser.add_argument("-jitlog", metavar = "jitlog_file", dest = "jitlog_file", nargs = 1, help = "use jitlog data")
    parser.add_argument("-pypylog", metavar = "pypylog_file", dest = "pypylog_file", nargs = 1, help = "use pypylog data")
    parser.add_argument("-run", metavar = "python_file args", dest = "python_file", nargs = "+", help = "run vmprof and convert profile")
    parser.add_argument("--nobrowser", action = "store_false", dest = "browser", default = True, help = "dont open firefox profiler")
    parser.add_argument("--zip", action = "store_true", dest = "zip", default = False, help = "save profile as sharable zip file")
    parser.add_argument("--nonative", action = "store_false", dest = "native", default = True, help = "disable native profiling")
    parser.add_argument("--nolines", action = "store_false", dest = "lines", default = True, help = "disable line profiling")

    args = parser.parse_args()

    zip_path = None
    pypylogpath = None
    jitlogpath = None
    times = None
    path_dict = None

    if args.python_file:
        path, jitlogpath, pypylogpath, times = run_vmprof(args.python_file[0], args.python_file[1:], args.native, args.lines)
        if jitlogpath is not None:
            jitlogpath = os.path.abspath(jitlogpath)
        if pypylogpath is not None:
            pypylogpath = os.path.abspath(pypylogpath)
        if args.zip:
            zip_path = path.replace(".prof", ".zip")
    elif args.convert_file:
        if ".zip" in args.convert_file[0]:
            extracted_folder = "tmp"
            zip_dict = load_zip_dict(args.convert_file[0], extracted_folder)
            if "times" in zip_dict:
                times = zip_dict["times"]
                zip_dict.pop("times")
            path_dict = extract_files(zip_dict, args.convert_file[0],  extracted_folder)
            path, jitlogpath, pypylogpath = get_paths(path_dict)
            atexit.register(cleanup, extracted_folder, path_dict)
        else:
            path = args.convert_file[0]
            if args.zip:
                zip_path = path.replace(".prof", ".zip")
            if args.jitlog_file:
                jitlogpath = os.path.abspath(args.jitlog_file[0])
            if args.pypylog_file:
                pypylogpath = os.path.abspath(args.pypylog_file[0])
    else:
        print("You must provide arguments")
        print("Either convert some profile with: python -m vmprofconvert -convert some_file.prof ...")
        print("Or run some code with: python -m vmprofconvert -run some_file.py <args>")
        exit(0)

    abspath = os.path.abspath(path)

    url = "https://profiler.firefox.com/from-url/http%3A%2F%2F127.0.0.1%3A5000%2Fprofile"
    url += "/?symbolServer=http%3A%2F%2F127.0.0.1%3A5000%2F"

    with open(abspath + ".json", "w") as output_file:
        json_profile, new_path_dict = convert_stats_with_pypylog(abspath, pypylogpath, times)
        output_file.write(json.dumps(json.loads(json_profile), indent=2))
        if args.zip:
            new_path_dict["prof"] = abspath
            if jitlogpath:
                new_path_dict["jitlog"] = jitlogpath
            if pypylogpath:
                new_path_dict["pypylog"] = pypylogpath
                new_path_dict["times"] = list(times)
            save_zip(zip_path, new_path_dict)
        if pypylogpath and os.path.exists(pypylogpath) and not args.convert_file:
            os.remove(pypylogpath)
    if args.browser:
        webbrowser.open(url, new=0, autoraise=True)
        start_server(abspath + ".json", jitlogpath, path_dict)


if __name__ == "__main__":
    main()



import os 
import json
import dis
import re
from jitlog.parser import parse_jitlog
from flask import Flask
from flask import request
from flask_cors import CORS, cross_origin

flaskapp = Flask(__name__)
cors = CORS(flaskapp)
flaskapp.config["CORS_HEADERS"] = "Content-Type"   

codeobj_dict = {}
jitlog_forest = {}

@flaskapp.get("/profile")
def getprofile():
    if os.path.exists(profilepath):
        return json.dumps(open(profilepath, "r").read())
    else:
        return ""
    
@flaskapp.post("/source/v1")
def source():
    response = {
        "symbolsLastModified": None,
        "sourceLastModified": None,
        "file": "",
        "source": "" 
    }   
    if request.data is not None:
        jsonobj = json.loads(request.data)
        response["file"] = jsonobj["file"]
        file = response["file"]
        if path_dict is not None:
            if file in path_dict:
                file = path_dict[file]
        if file is not None:
            if os.path.exists(file):
                with open(file, "r") as file:
                    response["source"] = file.read()
                return json.dumps(response)
    return response

@flaskapp.post("/asm/v1")
def asm():
    response = {
        "startAddress": "0x7",
        "size": "0x17",
        "arch": "arch", 
        "syntax": "jitlog",
        "instructions": []
    }
    if request.data is not None:
        jsonobj = json.loads(request.data)
        addr = jsonobj["startAddress"]
        if isinstance(addr, str) and addr != "0x-1":
            addr = int(addr, 16)
        response["startAddress"] = "0x0"
        funcname = jsonobj["debugName"]
        code = get_advanced_code(jitlogpath, addr, funcname)
        if len(code) != 0:
            response["instructions"] = code
            response["size"] = len(code)
    return json.dumps(response)

def get_code_object(path, path_dict):
    if path_dict is not None:
        if path in path_dict:
            path = path_dict[path]
    if path in codeobj_dict:
        return codeobj_dict[path]
    else:
        #maybe add limit
        with open(path, "rb") as file:
            content = file.read()
        codeobj_dict[path] = compile(content, path, "exec")
        return codeobj_dict[path]

def get_advanced_code(jitpath, addr, funcname):
    code = {}
    if jitpath is None or not os.path.exists(jitpath):
        return []
    if jitpath not in jitlog_forest:
        jitlog_forest[jitpath] = parse_jitlog(jitpath)
    forest = jitlog_forest[jitpath]
    trace = forest.get_trace_by_addr(addr)
    if trace is None:
        return []
    mp_data = get_mp_data(trace)
    ir_code = get_ir_code(trace.stages["opt"])
    code["pre"] = ir_code[0]# ir code from before first py line 
    for i, ir in enumerate(mp_data):
        if ir[2] != funcname:# sometimes there is all the ir code in a single asm frame
            continue
        key = ir[0] + str(ir[1])
        py_line = get_sourceline(ir[0], ir[1], path_dict).replace("\n", "")
        py_line += "  #" + ir[0] + ":" + str(ir[1])
        if py_line is not None:
            codeobject = get_code_object(ir[0], path_dict)
            bc_instr = get_bc_instruction(codeobject, ir[2], ir[1], ir[3])
            bc_str = bc_to_str(bc_instr)
            insert_code(code, key, py_line, bc_str, ir_code[i + 1])
    return code_dict_to_list(code)

def get_ir_code(stage_opt):
    ir_code = []
    indexes = [0]
    [indexes.append(mp.index) for mp in stage_opt.get_merge_points()]
    indexes.append(len(stage_opt.get_ops()))
    for i in range(len(indexes) - 1):
        ir_code.append(stage_opt.get_ops()[indexes[i]:indexes[i + 1]])
    return ir_code
   
def insert_code(code, key, py_line, bc_instr, ir_instr):
    if bc_instr is not None:
        if key not in code:
            code[key] = {
                "py_line": py_line.strip(),
                "bc": [{
                    "bc_line": bc_instr,
                    "ir_code": ir_instr 
                    }
                ]
            }
        else:
            nd = {
                "bc_line": bc_instr,
                "ir_code": ir_instr 
            }
            code[key]["bc"].append(nd)

def bc_to_str(bc_instr):
    bc = re.sub(" +", " ", bc_instr._disassemble())
    while not bc[0].isalpha():
        bc = bc[1:]
    return bc.strip()

def code_dict_to_list(code):
    index = 0
    lcode = []
    if "pre" in code:
        for ir_line in code["pre"]:
            lcode.append([index, "    " + ir_to_str(ir_line)])
            index += 1
        code.pop("pre")
    for v in code.keys():
        tmp = code[v]
        lcode.append([index, tmp["py_line"]])
        index += 1
        for bc in tmp["bc"]:
          lcode.append([index, "  " + bc["bc_line"]])
          index += 1
          for ir in bc["ir_code"]:
              lcode.append([index, "    " + ir_to_str(ir)])
              index += 1
    return lcode

def ir_to_str(ir):
    return str(ir).removeprefix("? = ")

def get_jitlog_ir(jitpath, addr):
    asm = []
    if jitpath is None or not os.path.exists(jitpath):
        return asm
    forest = parse_jitlog(jitpath)
    trace = forest.get_trace_by_addr(addr)
    if trace is not None:
        if "opt" in trace.stages:
            for i, op in enumerate(trace.stages["opt"].get_ops()):
                asm.append([i, str(op)])
    return asm

def get_mp_data(trace):
    offsets = []
    if trace is not None:
        if "opt" in trace.stages:
            mergepoints = trace.stages["opt"].get_merge_points()
            for mp in mergepoints:
                ofs = (mp.values[1], mp.values[2], mp.values[8], mp.values[4])# file, lineno, func, bc_offset
                offsets.append(ofs)
    return offsets

def get_all_bytecodes_rec(code):
    yield code
    for const in code.co_consts:
        if type(const) is not type(code):
            continue
        yield from get_all_bytecodes_rec(const)

def search_bytecode(module_code, funcname, linenumber):
    for code in get_all_bytecodes_rec(module_code):
        if code.co_name == funcname:# and code.co_firstlineno == linenumber:
            yield code

def get_bc_instruction(codeobject, funcname, linenumber, offset):
    l = list(search_bytecode(codeobject, funcname, linenumber))
    if len(l) != 1:
        return None
    bc = dis.Bytecode(l[0])
    for inst in bc:
        if inst.offset == offset:
            return inst
    
def get_sourceline(path, line, path_dict):
    if path_dict is not None:
        if path in path_dict:
            path = path_dict[path]
    if path is None or not os.path.exists(path):
        return None
    sourcelines = []
    line -= 1# line in debug_merge_points starts with 1
    with open(path, "r") as file:
        sourcelines = file.readlines()
    if len(sourcelines) > line:
        if sourcelines[line].strip() == "":
            return None
        return sourcelines[line]
    return None

def start_server(jsonpath, jitlog, pathdict):
    global profilepath , jitlogpath, path_dict
    profilepath = jsonpath
    jitlogpath = jitlog
    path_dict = pathdict
    flaskapp.run() 


import vmprof
import json
import os
from vmprof.reader import AssemblerCode, JittedCode
from vmprofconvert.processedformat import check_processed_profile
from vmprofconvert.pypylog import parse_pypylog, cut_pypylog, rescale_pypylog, filter_top_level_logs

CATEGORY_PYTHON = 0
CATEGORY_MEMORY = 1
CATEGORY_NATIVE = 2
CATEGORY_JIT = 3
CATEGORY_ASM = 4
CATEGORY_JIT_INLINED = 5
CATEGORY_MIXED = 6
CATEGORY_GC = 7
CATEGORY_INTERPRETER = 8

PPL_TIME = 0
PPL_ACTION = 1
PPL_STARTING = 2
PPL_DEPTH = 3

def convert(path):
    stats = vmprof.read_profile(path)
    c = Converter()
    c.walk_samples(stats)
    return c# return converter instance for testing
    
def convert_vmprof(path):
    c = Converter()
    stats = vmprof.read_profile(path)
    c.walk_samples(stats)
    return c# return converter instance for testing

def convert_stats(path):
    # new function because dumps_vmprof needs stats
    c = Converter()
    stats = vmprof.read_profile(path)
    c.walk_samples(stats)
    return c.dumps_vmprof(stats)

def convert_stats_with_pypylog(vmprof_path, pypylog_path, times):
    #times for cutting of logs after sampling ended
    c = Converter()
    stats = vmprof.read_profile(vmprof_path)
    c.walk_samples(stats)
    if pypylog_path  :
        pypylog = parse_pypylog(pypylog_path)
        if times is not None:
            total_runtime_micros = (times[1] - times[0]) * 1000000
            pypylog = cut_pypylog(pypylog, total_runtime_micros, stats.get_runtime_in_microseconds())
        pypylog = rescale_pypylog(pypylog, stats.get_runtime_in_microseconds())
        c.walk_pypylog(pypylog)
    return c.dumps_vmprof(stats), c.create_path_dict()# json_profile, file path dict

class Converter:
    def __init__(self):
        self.threads = {}
        self.counters = []
        self.libs = []# list of [name, debugname]
        self.libs_positions = {}# key is string
    
    def add_lib(self, name, debugname):
        if name == "-":
            return -1
        key = (name, debugname)
        if key in self.libs_positions:
            return self.libs_positions[key]
        else:
            libs_index = len(self.libs)
            self.libs.append([name, debugname])
            self.libs_positions[key] = libs_index
            return libs_index
        
    def create_path_dict(self):
        path_dict = {}
        for lib in self.libs:
            name, _ = lib
            if os.path.exists(name):
                path_dict[name] = os.path.abspath(name)
        return path_dict
    
    def create_pypylog_marker(self, pypylog, tid):
        self.threads[tid].create_pypylog_marker(pypylog)

    def get_unused_tid(self):
        if len(self.threads) == 0:
            return 7
        lowest_tid = min(list(self.threads.keys())) - 1
        return lowest_tid
    
    def walk_pypylog(self, pypylog):
        tid = self.get_unused_tid()
        plthread = None
        if tid not in self.threads:
            plthread = self.threads[tid] = Thread()
            plthread.name = "PyPyLog"
            plthread.tid = tid
        if plthread and pypylog:
            ppl_stack = []
            last_log = None
            last_close_time = -1
            mdiff = pypylog[1][PPL_TIME] - pypylog[0][PPL_TIME]# minimal time interval to squeeze an interpreter sample in between 
            for i in range(len(pypylog)):
                log = pypylog[i]
                if log[PPL_STARTING]:
                    ppl_stack.append(log)
                else:
                    ppl_stack.append(log)
                    if last_log[PPL_ACTION] == log[PPL_ACTION]: # Only top level actions wanted e.g. A[ B[ B]A] => Sample: A->B not A->B, A
                        if len(plthread.samples) >= 2 and ppl_stack[-2][PPL_TIME] - last_close_time >= mdiff:# dont add interp sample at start
                            self.add_pypylog_interp_sample(plthread, last_close_time + 1, ppl_stack[-2][PPL_TIME] - 1)
                            plthread.create_single_pypylog_interpreter_marker(last_close_time + 1, ppl_stack[-2][PPL_TIME] - 1)
                        self.add_pypylog_sample_from_stack(plthread, ppl_stack)
                        plthread.create_single_pypylog_marker(ppl_stack[-2], ppl_stack[-1])
                    last_close_time = log[PPL_TIME]
                    ppl_stack.pop()
                    ppl_stack.pop()
                last_log = log
            
    def add_pypylog_sample_from_stack(self, thread, stack_list):
        frames = []
        categories = []
        for log in stack_list[:-1]:# last frame is double (open + closed)
            if "gc" in log[PPL_ACTION]:
                categories.append(CATEGORY_GC)
            elif "jit" in log[PPL_ACTION]:
                categories.append(CATEGORY_JIT)
            frames.append(thread.add_frame(log[PPL_ACTION], -1, "", categories[-1], -1, -1))
        stackindex = thread.add_stack(frames, categories)
        start_time = stack_list[-2][PPL_TIME]
        end_time =  stack_list[-1][PPL_TIME]
        thread.add_sample(stackindex, start_time)
        thread.add_sample(stackindex, end_time) 

    def add_pypylog_sample(self, thread, logname, logtime_start, logtime_end):
        if "gc" in logname:
            category = CATEGORY_GC
        elif "jit" in logname:
            category = CATEGORY_JIT
        frameindex = thread.add_frame(logname, -1, "", category, -1, -1)
        stackindex = thread.add_stack([frameindex], [category])
        thread.add_sample(stackindex, logtime_start)
        thread.add_sample(stackindex, logtime_end)

    def add_pypylog_interp_sample(self, thread, logtime_start, logtime_end):
        frameindex = thread.add_frame("interp", -1, "", CATEGORY_INTERPRETER, -1, -1)
        stackindex = thread.add_stack([frameindex], [CATEGORY_INTERPRETER])
        thread.add_sample(stackindex, logtime_start)
        thread.add_sample(stackindex, logtime_end)

    
    def walk_samples(self, stats):
        sampletime = stats.end_time.timestamp() * 1000 - stats.start_time.timestamp() * 1000
        sampletime /= len(stats.profiles)

        if "start_time_offset" in stats.meta: # No version in stats TODO: Replace with version check when vmprof supports it
            sampletime = float(stats.getmeta("start_time_offset", "0")) * 1000

        category_dict = {}
        category_dict["py"] = CATEGORY_PYTHON
        category_dict["n"] = CATEGORY_NATIVE
        for i, sample in enumerate(stats.profiles):
            frames = []
            categories = []
            stack_info, _, tid, memory = sample
            if tid in self.threads:
                thread = self.threads[tid]
            else:
                thread = self.threads[tid] = Thread() 
                thread.tid = tid
                thread.name = "Thread " + str(len(self.threads) - 1)# Threads seem to need different names
            if stats.profile_lines:
                indexes = range(0, len(stack_info), 2)
            else:
                indexes = range(len(stack_info))

            for j in indexes:
                addr_info = stats.get_addr_info(stack_info[j])
                #remove jit frames # quick fix 
                if len(categories) != 0:
                    if not isinstance(stack_info[j], AssemblerCode) and categories[-1] == CATEGORY_JIT:
                        frames.pop()
                        categories.pop()
                if isinstance(stack_info[j], JittedCode):
                    frames.append(self.add_jit_frame(thread, categories, addr_info, frames))
                elif isinstance(stack_info[j], AssemblerCode):
                    self.check_asm_frame(categories, stack_info[j], thread, frames[-1])
                elif addr_info is None: # Class NativeCode isnt used
                    #pass
                    categories.append(CATEGORY_NATIVE)
                    frames.append(self.add_native_frame(thread, stack_info[j]))                   
                elif isinstance(stack_info[j], int): 
                    categories.append(category_dict[addr_info[0]])  
                    frames.append(self.add_vmprof_frame(addr_info, thread, stack_info, stats.profile_lines,categories[-1], j))
                   
            stackindex = thread.add_stack(frames, categories)
            timestamp = i * sampletime
            if "start_time_offset" in stats.meta: 
                timestamp = 1000 * stats.profiles[i][1] - sampletime# timestamp field in new version  
            thread.add_sample(stackindex, timestamp)
            if stats.profile_memory == True:
                self.counters.append([timestamp, memory * 1000])

    def add_vmprof_frame(self, addr_info, thread, stack_info, lineprof, category, j):# native or python frame
        funcname = addr_info[1]
        funcline = addr_info[2]
        filename = addr_info[3]
        lib_index = self.add_lib(filename, funcname)
        if lineprof:
            return thread.add_frame(funcname, int(-1 * stack_info[j + 1]), filename, category, lib_index, -1)# vmprof python line indexes are negative
        else:
            return thread.add_frame(funcname, funcline, filename, category, lib_index, -1)

    def add_jit_frame(self, thread, categories, addr_info, frames):
        funcname = addr_info[1]
        filename = addr_info[3]
        last_funcname, last_filename = self.get_last_func_file(thread, frames) 
        
        if len(categories) > 0 and categories[-1] == 0 and last_filename == filename and last_funcname == funcname:# if last frame is py and current is jit and both have the same function => replace with mixed frame
            frames.pop()
            categories.pop()
            categories.append(CATEGORY_MIXED)
        else:
            categories.append(CATEGORY_JIT)
        if addr_info is not None and int(addr_info[2]) >= 0:
            lib_index = self.add_lib(filename, funcname)
            return thread.add_frame(funcname, int(addr_info[2]), filename, categories[-1], lib_index, -1)# vmprof jit line indexes are positive
        else:
            return thread.add_frame(funcname, -1, filename, categories[-1], -1, -1)

    def add_native_frame(self, thread, stack_info):
        funcname = stack_info
        filename = ""
        frameindex = thread.add_frame(funcname, -1, filename, CATEGORY_NATIVE, -1, -1)
        return frameindex
    
    def check_asm_frame(self, categories, stack_info, thread, last_frame):
        if len(categories) > 0 and categories[-1] == 3:# if last frame is jit and current is asm => replace with inline jit frame
            categories.pop()
            categories.append(CATEGORY_JIT_INLINED)
            last_nativesymbol_index = thread.frametable[last_frame][1]
            thread.nativesymbols[last_nativesymbol_index][2] = stack_info
        else:# asm disabled
            pass
            #categories.append(CATEGORY_ASM)#asm
            #frames.append(thread.add_frame(stack_info[j], -1, ""))

    def get_last_func_file(self, thread, frames):
        if len(frames) == 0:
            return "", ""
        last_func_index = thread.frametable[frames[-1]][0]
        last_funcname_index = thread.functable[last_func_index][0]
        last_filename_index = thread.functable[last_func_index][1]
        last_funcname = thread.stringarray[last_funcname_index]
        last_filename = thread.stringarray[last_filename_index]
        return last_funcname, last_filename

    def dumps_static(self):
        processed_profile = {}
        processed_profile["meta"] = self.dump_static_meta()
        processed_profile["libs"] = []
        processed_profile["pages"] = []
        processed_profile["counters"] = []
        processed_profile["threads"] = self.dump_threads()
        check_processed_profile(processed_profile)
        return json.dumps(processed_profile)
    
    def dumps_vmprof(self, stats):
        processed_profile = {}
        processed_profile["meta"] = self.dump_vmprof_meta(stats)
        processed_profile["libs"] = self.dump_libs()
        processed_profile["pages"] = []
        if(stats.profile_memory):
            processed_profile["counters"] = [self.dump_counters()]
        else:
            processed_profile["counters"] = []
        processed_profile["threads"] = self.dump_threads()
        check_processed_profile(processed_profile)
        return json.dumps(processed_profile)
    
    def dump_libs(self):
        liblist = []
        for lib in self.libs:
            name, debugname = lib
            liblist.append(
                {
                    "name": name,
                    "path": name,
                    "debugName": debugname,
                    "debugPath": name,
                    "arch": ""
                }
            )
        return liblist
    
    def dump_threads(self):
        return [thread.dump_thread()for thread in list(self.threads.values())] 
    
    def dump_static_meta(self):
        static_meta = {}
        static_meta["interval"] = 0.4
        static_meta["startTime"] = 1477063882018.4387
        static_meta["abi"] = "x86_64-gcc3"
        static_meta["oscpu"] = "Intel Mac OS X 10.12"
        static_meta["platform"] = "Macintosh"
        static_meta["processType"] = 0
        static_meta["stackwalk"] = 1
        static_meta["debug"] = False
        static_meta["version"] = 27
        static_meta["importedFrom"] = "VMProf"
        static_meta["categories"] = self.dump_categories()
        static_meta["preprocessedProfileVersion"] = 47
        static_meta["symbolicated"] = True
        static_meta["markerSchema"] = []
        return static_meta

    def dump_vmprof_meta(self, stats):
        vmprof_meta = {}
        ms_for_sample = int(stats.get_runtime_in_microseconds() / len(stats.profiles))
        
        vmprof_meta["interval"] = ms_for_sample * 0.000001# seconds
        vmprof_meta["startTime"] = stats.start_time.timestamp() * 1000
        vmprof_meta["shutdownTime"] = stats.end_time.timestamp() * 1000
        vmprof_meta["abi"] = stats.interp # interpreter

        os   = stats.getmeta("os","default os")
        bits = stats.getmeta("bits","64")
        osdict = {"linux": "x11", "win64": "Windows", "win32": "Windows", "mac os x": "Macintosh"}
        
        vmprof_meta["oscpu"] = f"{osdict[os]} {bits}bit"
        vmprof_meta["platform"] = osdict[os]
        vmprof_meta["processType"] = 0
        vmprof_meta["stackwalk"] = 1
        vmprof_meta["debug"] = False
        vmprof_meta["version"] = 27
        vmprof_meta["importedFrom"] = "VMProf"
        vmprof_meta["categories"] = self.dump_categories()
        vmprof_meta["preprocessedProfileVersion"] = 47
        vmprof_meta["symbolicated"] = True
        vmprof_meta["markerSchema"] = []
        
        return vmprof_meta
    
    def dump_categories(self):
        categories = []
        categories.append(
            {
                "name": "Python",
                "color": "yellow",
                "subcategories": [
                    "Other"
                ]
            }
        )
        categories.append(
            {
                "name": "Memory",
                "color": "red",
                "subcategories": [
                    "Other"
                ]
            }
        )    
        categories.append(
            {
                "name": "Native",
                "color": "lightblue",
                "subcategories": [
                    "Other"
                ]
            }
        )    
        categories.append(
            {
                "name": "JIT",
                "color": "purple",
                "subcategories": [
                    "Other"
                ]
            }
        )    
        categories.append(
            {
                "name": "ASM",
                "color": "blue",
                "subcategories": [
                    "Other"
                ]
            }
        )
        categories.append(
            {
                "name": "JIT(Inlined)",
                "color": "purple",
                "subcategories": [
                    "Other"
                ]
            }
        )   
        categories.append(
            {
                "name": "Mixed",
                "color": "orange",
                "subcategories": [
                    "Other"
                ]
            }
        )
        categories.append(
            {
                "name": "Garbage Collection",
                "color": "green",
                "subcategories": [
                    "Other"
                ]
            }
        )
        categories.append(
            {
                "name": "Interpreter",
                "color": "yellow",
                "subcategories": [
                    "Other"
                ]
            }
        )
        return categories
           
    def dump_counters(self):
        counter = {}
        counter["name"] = "Memory"
        counter["category"] = "Memory"
        counter["description"] = "Amount of allocated memory"
        counter["pid"] = "51580"
        counter["mainThreadIndex"] = 0
        memory_in_alloc_form =  self.get_mem_allocations()
        counter["sampleGroups"] = [
            {
                "id": 0,
                "samples": {
                     "length": len(memory_in_alloc_form),
                     "time": [mem[0] for mem in memory_in_alloc_form],
                     "count": [mem[1] for mem in memory_in_alloc_form]
                }
            }
        ]
        return counter
    
    def get_mem_allocations(self):
        # Firefox Profiler seems to need two non zero samples
        mem_diff = [self.counters[0],self.counters[0]]
        current_mem = mem_diff[0][1]
        for ctr in self.counters[1:len(self.counters) - 1 ]:
            mem_diff.append([ctr[0], (ctr[1] - current_mem)])
            current_mem = ctr[1]
        return mem_diff


class Thread:
    def __init__(self):
        self.stringarray = []
        self.stringarray_positions = {}
        self.stacktable = []# list of [frameindex, stacktableindex_or_None, category] cat{0 = py, 1 = mem, 2 = native, 3 = jit, 4 = asm, 5 =  jit_inline, 6 = mixed}
        self.stacktable_positions = {}
        self.functable = []# list of [stringtable_index, stringtable_index, int, resource_index] funcname, filename, line  line == -1 if profile_lines == False, resource_index
        self.funtable_positions = {}
        self.frametable = []# list of [functable_index, nativesymbol_index, line]   
        self.frametable_positions = {}# key is string
        self.samples = []# list of [stackindex, time in ms], no need for sample_positions
        self.nativesymbols = []# list of [libindex, stringindex, addr]
        self.nativesymbols_positions = {}# key is (libindex, string)
        self.resourcetable = []# list of [libindex, stringindex]
        self.resourcetable_positions = {}# key is (libindex, stringindex)
        self.markers = []# list of [startime, endtime, stringindex]

    def create_pypylog_marker(self, pypylog):
        interperter_string_id = self.add_string("interpreter")
        for i in range(int(len(pypylog)/2)):
            start_log = pypylog[2*i]
            stop_log = pypylog[2*i+1]
            starttime = start_log[0]
            endtime = stop_log[0]
            name = start_log[1]
            st_id = self.add_string(name)
            self.add_marker(starttime, endtime, st_id)
            if i < ((len(pypylog)/2) - 2):
                next_log = pypylog[2 * i + 2]
                next_logtime_start = next_log[0]
                if abs(endtime - next_logtime_start) > 2:
                    self.add_marker(endtime + 1, next_logtime_start - 1, interperter_string_id)

    def create_single_pypylog_marker(self, start_log, stop_log):
        starttime = start_log[PPL_TIME]
        endtime = stop_log[PPL_TIME]
        name = start_log[PPL_ACTION]
        st_id = self.add_string(name)
        self.add_marker(starttime, endtime, st_id)

    def create_single_pypylog_interpreter_marker(self, starttime, endtime):
        st_id = self.add_string("interpreter")## move out
        self.add_marker(starttime, endtime, st_id)
            
    def add_marker(self, starttime, endtime, stringtable_index):
        self.markers.append([starttime, endtime, stringtable_index])

    def add_string(self, string):
        if string in self.stringarray_positions:
            return self.stringarray_positions[string]
        else:
            result = len(self.stringarray)
            self.stringarray.append(string)
            self.stringarray_positions[string] = result
            return result
        
    def add_stack(self, stack, categories):
        #stack is a list of frametable indexes
        if not stack:
            return None
        else:
            top = stack[-1]
            rest = stack[:-1]
            top_category = categories[-1]
            rest_categories = categories[:-1]
            rest_index = self.add_stack(rest, rest_categories)
            key = (top, rest_index, top_category)
            if key in self.stacktable_positions:
                return self.stacktable_positions[key]
            else:
                result = len(self.stacktable)
                self.stacktable.append([top, rest_index, top_category])
                self.stacktable_positions[key] = result
                return result
            
    def add_func(self, func, file, line, category, libindex):
        key = (func, file, line, category)
        if key in self.funtable_positions:
            return self.funtable_positions[key]
        else:
            stringtable_index_func = self.add_string(func)
            stringtable_index_file = self.add_string(file)
            resource_index = self.add_resource(libindex, stringtable_index_func)
            result = len(self.functable)
            is_python = category in (CATEGORY_PYTHON, CATEGORY_MIXED, CATEGORY_JIT, CATEGORY_JIT_INLINED)
            self.functable.append([stringtable_index_func, stringtable_index_file, line, resource_index, is_python])
            self.funtable_positions[key] = result
            return result
            
    def add_frame(self, funcname, line, file, category, libindex, addr):
        key = (funcname, line, category)
        if key in self.frametable_positions:
            return self.frametable_positions[key]
        else:
            functable_index = self.add_func(funcname, file, line, category, libindex)
            frametable_index = len(self.frametable)
            nativesymbol_index = self.add_nativesymbol(libindex, funcname, addr)
            self.frametable.append([functable_index, nativesymbol_index, line, category])
            self.frametable_positions[key] = frametable_index
            return frametable_index
        
    def add_sample(self, stackindex, time):
        self.samples.append([stackindex, time]) # stackindex, ms since starttime
    
    def add_nativesymbol(self, libindex, funcname, addr):
        if libindex == -1:
            return -1
        key = (libindex, funcname, addr)
        if key in self.nativesymbols_positions:
            return self.nativesymbols_positions[key]
        else:
            funcname_index = self.add_string(funcname)
            nativesymbol_index = len(self.nativesymbols)
            self.nativesymbols.append([libindex, funcname_index, addr])
            self.nativesymbols_positions[key] = nativesymbol_index
            return nativesymbol_index
    
    def add_resource(self, libindex, stringindex):
        if libindex == -1:
            return -1
        key = (libindex, stringindex)
        if key in self.resourcetable_positions:
            return self.resourcetable_positions[key]
        else:
            resource_index = len(self.resourcetable)
            self.resourcetable.append([libindex, stringindex])
            self.resourcetable_positions[key] = resource_index
            return resource_index
    
    def dump_thread(self):
        thread = {}
        thread["name"] = self.name
        thread["isMainThread"] = False
        thread["processType"] = "default"
        thread["processName"] = "Parent Process"
        thread["processStartupTime"] = 0
        thread["processShutdownTime"] = None
        thread["registerTime"] = 23.841461000000002
        thread["unregisterTime"] = None
        thread["tid"] = self.tid
        thread["pid"] = "51580"
        thread["markers"] = self.dump_markers()
        thread["nativeSymbols"] = self.dump_nativesymbols()
        thread["frameTable"] = self.dump_frametable()
        thread["funcTable"] = self.dump_functable()
        thread["resourceTable"] = self.dump_resourcetable()
        thread["stackTable"] = self.dump_stacktable()
        thread["samples"] = self.dump_samples()
        thread["stringArray"] = self.dump_stringarray()
        return thread
    
    def dump_markers(self):
        markers = {}
        markers["data"] = [{"type": "PyPyLog"} for _ in self.markers]
        markers["name"] = [m[2] for m in self.markers]
        markers["startTime"] = [m[0] for m in self.markers]
        markers["endTime"] = [m[1] for m in self.markers]
        markers["phase"] = [1 for _ in self.markers]
        markers["category"] = [7 for _ in self.markers]
        markers["length"] = len(self.markers)
        return markers
 
    def dump_resourcetable(self):
        resourcetable = {}
        resourcetable["lib"] = [nsym[0] for nsym in self.resourcetable]
        resourcetable["name"] = [nsym[1] for nsym in self.resourcetable]
        resourcetable["host"] = [None for _ in self.resourcetable]
        resourcetable["type"] = [1 for _ in self.resourcetable]
        resourcetable["length"] = len(self.resourcetable)
        return resourcetable
    
    def dump_nativesymbols(self):
        nativesymbols = {}
        nativesymbols["libIndex"] = [nsym[0]for nsym in self.nativesymbols]
        nativesymbols["address"] = [nsym[2] for nsym in self.nativesymbols]
        nativesymbols["name"] = [nsym[1]for nsym in self.nativesymbols]
        nativesymbols["functionSize"] = [None for _ in self.nativesymbols]
        nativesymbols["length"] = len(self.nativesymbols)
        return nativesymbols

    def dump_frametable(self):
        ftable = {}
        ftable["address"] = [-1 for _ in self.frametable]
        ftable["inlineDepth"] = [0 for _ in self.frametable]
        ftable["category"] = [frame[3] for frame in self.frametable]
        ftable["subcategory"] = [None for _ in self.frametable]
        ftable["func"] = [frame[0] for frame in self.frametable]
        ftable["innerWindowID"] = [0 for _ in self.frametable]
        ftable["implementation"] = [None for frame in self.frametable]
        ftable["nativeSymbol"] = [frame[1] for frame in self.frametable]
        ftable["line"] = self.get_frametable_lines()
        ftable["length"] = len(self.frametable)
        return ftable
    
    def get_frametable_lines(self):
        lines = []
        for frame in self.frametable:
            if frame[2] == -1:
                lines.append(None)
            else:
                lines.append(int(frame[2]))
        return lines
    
    def dump_functable(self):
        ftable = {}
        ftable["isJS"] = [func[4] for func in self.functable]
        ftable["relevantForJS"] = [False for _ in self.functable]
        ftable["name"] = [func[0] for func in self.functable]
        ftable["resource"] = [func[3] for func in self.functable]
        linenumbers, filenames = self.get_processed_filelines()
        ftable["fileName"] = filenames
        ftable["lineNumber"] = linenumbers
        ftable["columnNumber"] = [None for _ in self.functable]
        ftable["length"] = len(self.functable)
        return ftable

    def dump_stacktable(self):
        stable = {}
        stable["frame"] = [stack[0] for stack in self.stacktable]
        stable["category"] = [stack[2] for stack in self.stacktable]
        stable["subcategory"] = [None for _ in self.stacktable]
        stable["prefix"] = [stack[1] for stack in self.stacktable]
        stable["length"] = len(self.stacktable)
        return stable
    
    def dump_samples(self):
        samples = {}
        samples["stack"] = [sample[0] for sample in self.samples]
        samples["time"] = [sample[1] for sample in self.samples]
        samples["eventDelay"] = [7 for _ in self.samples]
        samples["weightType"] = "samples"
        samples["weight"] = None
        samples["length"] = len(self.samples)
        return samples
    
    def dump_stringarray(self):
        return [str(string) for string in self.stringarray]
    
    def get_processed_filelines(self):
        linenumbers = []
        filenames = []
        for func in self.functable:
            if self.stringarray[func[1]] == "-":
                linenumbers.append(None)
            elif func[2] == -1:
                linenumbers.append(None)
            else:
                linenumbers.append(int(func[2]))

        for func in self.functable:
            if self.stringarray[func[1]] == "-":
                filenames.append(None)
            else:
                filenames.append(func[1])
        
        return linenumbers, filenames



import os
from jitlog.parser import parse_jitlog
from symbolserver import get_jitlog_ir , bc_to_str
from symbolserver import get_mp_data, get_sourceline, get_bc_instruction, insert_code, code_dict_to_list, get_ir_code, ir_to_str, get_code_object

class Dummyinstruction:
    def __init__(self, str):
        self.__str__ = str
    
    def _disassemble(self):
        return self.__str__

def test_get_jitlog_ir():
    path = os.path.join(os.path.dirname(__file__), "profiles/pypy-pystone.prof.jit")
    asm = get_jitlog_ir(path, 140682485974400)
    
    assert len(asm) == 44

def test_get_empty_jitlog_ir():
    path = os.path.join(os.path.dirname(__file__), "profiles/pypy-pystone.prof.jit")
    asm = get_jitlog_ir(path, 115)# no ir for addr 115

    assert asm == []

def test_get_mp_data():
    path = os.path.join(os.path.dirname(__file__), "profiles/pypy-pystone.prof.jit")
    forest = parse_jitlog(path)
    trace = forest.get_trace_by_addr(140682485974400)
    mp_data = get_mp_data(trace)

    assert len(mp_data) == 18

def test_get_bytecode():
    path = os.path.join(os.path.dirname(__file__), "profiles/bctest.py")
    with open(path, "rb") as file:
        content = file.read()
    code = compile(content, path, "exec")
    instr = get_bc_instruction(code, "get_name", 6, 0)

    assert instr.offset == 0
    assert instr.argval == "self"
    assert instr.opname == "LOAD_FAST"

def test_get_source_line():
    path = os.path.join(os.path.dirname(__file__), "profiles/bctest.py")
    line = get_sourceline(path, 4, {})# merge_point line starts at 1

    assert line.strip() == "self.name = \"Floppa\""

def test_insert_code():
    code = {}
    key = "example key"
    py_line0 = "example py_line"
    bc_instr0 = "example bytecode"
    ir_instr0 = ["example ir code"]
    py_line1 = "example py_line"
    bc_instr1 = "other example bytecode"
    ir_instr1 = ["other example ir code"]

    insert_code(code, key, py_line0, bc_instr0, ir_instr0)
    insert_code(code, key, py_line1, bc_instr1, ir_instr1)

    expected_code = {
        "py_line": "example py_line",
        "bc": [
            {
                "bc_line": "example bytecode",
                "ir_code": ["example ir code"]
            }, 
            {
                "bc_line": "other example bytecode",
                "ir_code": ["other example ir code"]
            }
        ]
    }

    assert code[key] == expected_code

def test_code_dict_to_list():
    codedict = {
        "first key": {
            "py_line":  "first python line",
            "bc": [
                {
                    "bc_line": "first bytecode line",
                    "ir_code": ["first ir line"]
                }
            ]
        },
        "second key": {
            "py_line": "second python line",
            "bc": [
                {
                    "bc_line": "second bytecode line",
                    "ir_code": []
                },
                {
                    "bc_line": "third bytecode line",
                    "ir_code": []
                }
            ]
        }
    }
    expected_code_list = [[0, "first python line"],
                          [1, "  first bytecode line"],
                          [2, "    first ir line"],
                          [3, "second python line"],
                          [4, "  second bytecode line"],
                          [5, "  third bytecode line"]]
    
    codelist = code_dict_to_list(codedict)
    
    assert codelist == expected_code_list

def test_get_ir_code():
    path = os.path.join(os.path.dirname(__file__), "profiles/pypy-pystone.prof.jit")
    forest = parse_jitlog(path)
    trace = forest.get_trace_by_addr(140682485974400)
    ir_code = get_ir_code(trace.stages["opt"])

    assert len(ir_code) == 19

def test_ir_to_str():
    path = os.path.join(os.path.dirname(__file__), "profiles/pypy-pystone.prof.jit")
    forest = parse_jitlog(path)
    trace = forest.get_trace_by_addr(140682485974400)
    ir_code = get_ir_code(trace.stages["opt"])
    ir_instr = ir_code[1][0]
    ir_string = ir_to_str(ir_instr)
    expected_ir_str = "guard_value(i4, 4, @<ResumeGuardDescr object at 0x177f580>)"

    assert ir_string == expected_ir_str

def test_bc_to_str():
    bc0 = Dummyinstruction("55 LOAD_FAST         2 (float)")
    bc1 = Dummyinstruction("55   >>>>>   CALL_FUNCTION         4 (started)")
    pretty_bc0 = bc_to_str(bc0)
    pretty_bc1 = bc_to_str(bc1)

    assert pretty_bc0 == "LOAD_FAST 2 (float)"
    assert pretty_bc1 == "CALL_FUNCTION 4 (started)"

def test_get_example_bytecode():
    jitpath = os.path.join(os.path.dirname(__file__), "profiles/example.jitlog")
    pypath = os.path.join(os.path.dirname(__file__), "profiles/example.py")
    forest = parse_jitlog(jitpath)
    trace = forest.get_trace_by_addr(140249103068976)
    mp_data = get_mp_data(trace)
    codeobject = get_code_object(pypath, {})
    bc_instr = get_bc_instruction(codeobject, mp_data[2][2], mp_data[2][1], mp_data[2][3])    

    assert mp_data[2] == ("example.py", 3, "function_a", 12)
    assert bc_instr.offset == 12
    assert bc_instr.starts_line == 3
    assert bc_instr.opname == "JUMP_ABSOLUTE"

def test_get_example_ir_code():
    jitpath = os.path.join(os.path.dirname(__file__), "profiles/example.jitlog")
    forest = parse_jitlog(jitpath)
    trace = forest.get_trace_by_addr(140249103068976)
    ir_code = get_ir_code(trace.stages["opt"])
    expected_ir_code_7 = ["? = setfield_gc(p11, i44, @<FieldS pypy.module.__builtin__.functional.W_IntRangeIterator.inst_current 8>)",
                        "? = guard_not_invalidated(, @<ResumeGuardDescr object at 0x7f8e48c949e0>)",
                        "i48 = getfield_raw_i(140249191372576, @<FieldS pypysig_long_struct.c_value 0>)",
                        "i50 = int_sub(i48, 1)",
                        "? = setfield_raw(140249191372576, i50, @<FieldS pypysig_long_struct.c_value 0>)",
                        "i54 = int_lt(i50, 0)",
                        "? = guard_false(i54, @<ResumeGuardDescr object at 0x7f8e48c94a40>)"]
    
    assert len(ir_code) == 9
    for i, ir in enumerate(ir_code[7]):
        assert str(ir) == expected_ir_code_7[i]

